#include <omp.h>

#include "swe.h"
#include "dist_array.h"


extern real_t* rank_time;
extern bool border_exchange_barrier;


// length: The number of elements in the buffer,
//         not the memory size of the buffer.
static void zero_memory(real_t *buffer, int_t length) {
    /* Simplest implementation for now */
    // memset(buffer, 0, size);

    #pragma omp parallel for
    for (int_t i = 0; i < length; i++) {
        buffer[i] = 0;
    }
}


void realloc_dist_array(dist_array_t *result, int_t gsize[2]) {
    result->global_size[0] = gsize[0];
    result->global_size[1] = gsize[1];

    result->origin[0] = 0;

    for (int grid_y = 0; grid_y < coords[0]; grid_y++) {
        result->origin[0] +=
            gsize[0] / dims[0] + ((grid_y < (gsize[0] % dims[0])) ? 1 : 0);
    }

    result->origin[1] = 0;

    for (int grid_x = 0; grid_x < coords[1]; grid_x++) {
        result->origin[1] +=
            gsize[1] / dims[1] + ((grid_x < (gsize[1] % dims[1])) ? 1 : 0);
    }

    result->size[0] = gsize[0];
    result->size[1] = gsize[1];

    real_t start = omp_get_wtime();

    for (int b = 0; b < result->n_buffers; b++) {
        result->data[b] =
            realloc(result->data[b],
                    (gsize[0] + 2 * result->border) *
                    (gsize[1] + 2 * result->border) *
                    sizeof(real_t));
    }

    real_t end = omp_get_wtime();
    rank_time[realloc_index] += end - start;
    start = end;

    for (int b = 0; b < result->n_buffers; b++) {
        zero_memory(result->data[b],
                    (gsize[0] + 2 * result->border) *
                    (gsize[1] + 2 * result->border));
    }

    end = omp_get_wtime();
    rank_time[memset_realloc_index] += end - start;
    start = end;

    // TODO: Can this be simplified? At least profile it to see if it matters.
    MPI_Type_free(&(result->column));
    MPI_Type_free(&(result->row));

    MPI_Type_vector(result->size[0], result->border,
                    result->size[1] + 2 * result->border, mpi_real_t,
                    &(result->column));

    MPI_Type_commit(&(result->column));

    MPI_Type_vector(result->border, result->size[1] + 2 * result->border,
                    result->size[1] + 2 * result->border, mpi_real_t,
                    &(result->row));

    MPI_Type_commit(&(result->row));
}


dist_array_t * malloc_dist_array(int_t gsize[2], int_t border, int_t n_buffers)
{
    dist_array_t* result = malloc(sizeof(dist_array_t));
    result->global_size[0] = gsize[0];
    result->global_size[1] = gsize[1];
    result->border = border;
    result->n_buffers = n_buffers;

    result->origin[0] = 0;

    for (int grid_y = 0; grid_y < coords[0]; grid_y++) {
        result->origin[0] +=
            gsize[0] / dims[0] + ((grid_y < (gsize[0] % dims[0])) ? 1 : 0);
    }

    result->origin[1] = 0;

    for (int grid_x = 0; grid_x < coords[1]; grid_x++) {
        result->origin[1] +=
            gsize[1] / dims[1] + ((grid_x < (gsize[1] % dims[1])) ? 1 : 0);
    }

    result->size[0] = gsize[0];
    result->size[1] = gsize[1];

    real_t start = omp_get_wtime();

    result->data = malloc(n_buffers * sizeof(real_t*));

    for (int b = 0; b < n_buffers; b++) {
        result->data[b] =
            malloc((gsize[0] + 2 * border) * (gsize[1] + 2 * border) *
                   sizeof(real_t));
    }

    real_t end = omp_get_wtime();
    rank_time[malloc_index] += end - start;
    start = end;

    for (int b = 0; b < n_buffers; b++) {
        zero_memory(result->data[b],
                    (gsize[0] + 2 * border) * (gsize[1] + 2 * border));
    }

    end = omp_get_wtime();
    rank_time[memset_malloc_index] += end - start;
    start = end;

    MPI_Type_vector(result->size[0],
                    border, result->size[1] + 2 * border, mpi_real_t,
                    &(result->column));

    MPI_Type_commit(&(result->column));

    MPI_Type_vector(border, result->size[1] + 2 * border,
                    result->size[1] + 2 * border, mpi_real_t, &(result->row));

    MPI_Type_commit(&(result->row));

    rank_time[mpi_type_index] += omp_get_wtime() - start;

    return result;
}


// t: Index of timestep buffer
void border_exchange(dist_array_t *a, int_t t) {
    if (border_exchange_barrier) {
        real_t start = omp_get_wtime();
        MPI_Barrier(cart);
        rank_time[wait_index] += omp_get_wtime() - start;
    }

    real_t start = omp_get_wtime();
    
    /* Stage 1: east/west */
    MPI_Sendrecv (
        &(DARRAY(a,t,0,0)), 1, a->column, west, 0,
        &(DARRAY(a,t,0,a->size[1])), 1, a->column, east, 0,
        cart, MPI_STATUS_IGNORE
    );
    MPI_Sendrecv (
        &(DARRAY(a,t,0,a->size[1]-a->border)), 1, a->column, east, 0,
        &(DARRAY(a,t,0,-a->border)), 1, a->column, west, 0,
        cart, MPI_STATUS_IGNORE
    );

    /* Stage 2: North/south */
    MPI_Sendrecv (
        &(DARRAY(a,t,0,-a->border)), 1, a->row, north, 0,
        &(DARRAY(a,t,a->size[0],-a->border)), 1, a->row, south, 0,
        cart, MPI_STATUS_IGNORE
    );

    MPI_Sendrecv (
        &(DARRAY(a,t,a->size[0]-a->border,-a->border)), 1, a->row, south, 0,
        &(DARRAY(a,t,-a->border,-a->border)), 1, a->row, north, 0,
        cart, MPI_STATUS_IGNORE
    );

    rank_time[rank_border_index] += omp_get_wtime() - start;
}


/* Rotate data buffers */
void
swap_buffers ( dist_array_t *a )
{
    // Swap the first and the last buffer.

    // The new first buffer needs the data from the old last buffer.
    // The new last buffer needs a memory region that is different from the new
    // first buffer, so we reuse the memory region from the old first buffer.

    // Any buffer between the first and the last is assumed to be a temporary
    // buffer which will be overwritten during the next iteration, so we don't
    // need to do anything about it here.

    real_t *tmp = a->data[0];
    a->data[0] = a->data[a->n_buffers - 1];
    a->data[a->n_buffers - 1] = tmp;
}


void
free_dist_array ( dist_array_t *old )
{
    MPI_Type_free ( &(old->column) );
    MPI_Type_free ( &(old->row) );
    for ( int b=0; b<old->n_buffers; b++ )
        free ( old->data[b] );
    free ( old->data );
    free ( old );
}
