#include <inttypes.h>
#include <omp.h>
#include <stdlib.h>

#include "grid.h"


extern real_t* rank_time;


void update_child_grid(grid_t *child, grid_t *parent, point_t origin_in_parent,
                       length_t length_in_parent, int_t refinement)
{
    child->origin_in_parent = origin_in_parent;
    child->length_in_parent = length_in_parent;

    child->origin.y = parent->origin.y + origin_in_parent.y / parent->total_refinement;
    child->origin.x = parent->origin.x + origin_in_parent.x / parent->total_refinement;

    child->length.y = (length_in_parent.y - 1) * refinement + 1;
    child->length.x = (length_in_parent.x - 1) * refinement + 1;

    child->global_end.y = child->origin.y + (child->length.y - refinement + 1)
                        / child->total_refinement;

    child->global_end.x = child->origin.x + (child->length.x - refinement + 1)
                        / child->total_refinement;

    int_t size[] = {child->length.y, child->length.x};

    realloc_dist_array(child->pn, size);
    realloc_dist_array(child->pnu, size);
    realloc_dist_array(child->pnv, size);
    realloc_dist_array(child->n, size);
    realloc_dist_array(child->u, size);
    realloc_dist_array(child->v, size);
    realloc_dist_array(child->pnuv, size);
    realloc_dist_array(child->du, size);
    realloc_dist_array(child->dv, size);

    initialize_child_grids(parent, child, refinement);
}


grid_t * new_child_grid(grid_t *parent, int_t refinement,
                        point_t origin_in_parent, length_t length_in_parent)
{
    grid_t *child = malloc(sizeof(grid_t));

    child->level = parent->level + 1;
    child->parent = parent;

    child->origin_in_parent = origin_in_parent;
    child->length_in_parent = length_in_parent;

    child->origin.y = parent->origin.y + origin_in_parent.y / parent->total_refinement;
    child->origin.x = parent->origin.x + origin_in_parent.x / parent->total_refinement;

    child->total_refinement = parent->total_refinement * refinement;

    // Length = How many elements does the grid hold in a given direction?
    // Length of child = The number of elements in child for each element in
    // parent (defined by refinement) for the first n-1 elements in parent,
    // plus only one additional element for the last element in parent, since
    // the child also stops there.

    // Example: parent.length = 4, refinement = 2

    // |   |   |   |
    // 0   1   2   3 = 4 elements in parent

    // | | | | | | |
    // 0 1 2 3 4 5 6 = 7 elements in child

    // For each of element 0, 1 and 2 in parent, child has 2 elements.
    // For element 3 in parent, child has only 1 element.
    child->length.y = (length_in_parent.y - 1) * refinement + 1;
    child->length.x = (length_in_parent.x - 1) * refinement + 1;

    child->global_end.y = child->origin.y + (child->length.y - refinement + 1)
                        / child->total_refinement;

    child->global_end.x = child->origin.x + (child->length.x - refinement + 1)
                        / child->total_refinement;

    child->dy = parent->dy / refinement;
    child->dx = parent->dx / refinement;
    child->dt = parent->dt / refinement;

    real_t start = omp_get_wtime();

    new_grid_common(child);

    rank_time[allocate_child_grid_index] += omp_get_wtime() - start;

    initialize_child_grids(parent, child, refinement);

    return child;
}


grid_t * new_root_grid(point_t origin, length_t length,
                       real_t dy, real_t dx, real_t dt) {
    grid_t *grid = malloc(sizeof(grid_t));

    grid->level = 0;
    grid->parent = NULL;

    grid->origin = origin;
    grid->length = length;

    grid->global_end.y = origin.y + length.y - 1;
    grid->global_end.x = origin.x + length.x - 1;

    grid->total_refinement = 1;

    grid->dy = dy;
    grid->dx = dx;
    grid->dt = dt;

    new_grid_common(grid);

    return grid;
}


void new_grid_common(grid_t *grid) {
    grid->n_children = 0;
    grid->children = malloc(0);
    grid->border = 1; // TODO: Not hardcode
    grid->regrid_ticks = 0;

    int_t size[] = {grid->length.y, grid->length.x};

    // Allocate the double-buffered arrays w. border/halo width 1
    grid->pn = malloc_dist_array(size, 1, 3);
    grid->pnu = malloc_dist_array(size, 1, 3);
    grid->pnv = malloc_dist_array(size, 1, 3);
    grid->n = malloc_dist_array(size, 1, 2);
    grid->u = malloc_dist_array(size, 1, 2);
    grid->v = malloc_dist_array(size, 1, 2);

    // Allocate single-buffered arrays w. border/halo width 1
    grid->pnuv = malloc_dist_array(size, 1, 1);
    grid->du = malloc_dist_array(size, 1, 1);
    grid->dv = malloc_dist_array(size, 1, 1);
}


void add_child_grid(grid_t *parent, point_t origin, length_t length) {
    grid_t *child = new_child_grid(parent, 2, origin, length);

    parent->children =
        realloc(parent->children, (parent->n_children + 1) * sizeof(grid_t *));

    parent->n_children++;
    parent->children[parent->n_children - 1] = child;
}


// Remove one child, but keep the rest
// void remove_child_grid(grid_t *child, grid_t *parent) {
//     free_grid(child);

//     parent->children =
//         realloc(parent->children, (parent->n_children - 1) * sizeof(grid_t *));

//     parent->n_children--;
// }


// t: Time step for child
// refinement: How much more zoomed in is the child relative to the parent.
// Assume that dt_p/dt_c == dx_p/dx_c == dy_p/dy_c
void set_grid_borders(grid_t *child, dist_array_t *a_p, dist_array_t *a_c,
                      int_t t, int_t refinement, int_t buffer)
{
    real_t start = omp_get_wtime();

    // Case: Parent surrounds child
    border_t borders[] = {NORTH, SOUTH, WEST, EAST};

    for (int_t i = 0; i < 4; i++) {
        set_grid_border(child, a_p, a_c, t, refinement, borders[i], buffer);
    }

    // TODO: More complicated cases?

    rank_time[grid_border_index] += omp_get_wtime() - start;
}


// Find the value of the child from the value of the parent by linear
// interpolation in both time and space.
// a_p:         Parent array
// a_c:         Child array
// t:           Time step for child
// refinement:  How much more zoomed in is the child relative to the parent.
// Assume that dt_p/dt_c == dx_p/dx_c == dy_p/dy_c
void set_grid_border(grid_t *child, dist_array_t *a_p, dist_array_t *a_c,
                     int_t t, int_t refinement, border_t border, int_t buffer) {
    // Only works for border size 1 at the moment

    int_t
        Y,
        X,
        B = a_p->n_buffers - 1; // Index of last buffer

    switch (border) {
        case NORTH:
            Y = child->origin_in_parent.y;
            X = child->origin_in_parent.x;

            # pragma omp parallel for
            for (int_t x = 0; x < child->length.x; x++) {
                DARRAY(a_c, buffer, -1, x) =
                    (DARRAY(a_p, 0, Y,   x/2   + X) * (refinement - t) * (refinement - x % refinement)
                   + DARRAY(a_p, B, Y,   x/2   + X) * t                * (refinement - x % refinement)
                   + DARRAY(a_p, 0, Y,   x/2+1 + X) * (refinement - t) * (x % refinement)
                   + DARRAY(a_p, B, Y,   x/2+1 + X) * t                * (x % refinement)
                   + DARRAY(a_p, 0, Y-1, x/2   + X) * (refinement - t) * (refinement - x % refinement)
                   + DARRAY(a_p, B, Y-1, x/2   + X) * t                * (refinement - x % refinement)
                   + DARRAY(a_p, 0, Y-1, x/2+1 + X) * (refinement - t) * (x % refinement)
                   + DARRAY(a_p, B, Y-1, x/2+1 + X) * t                * (x % refinement)
                   ) / (refinement * refinement * refinement); // In y, x and t direction
            }

            break;

        case SOUTH:
            Y = child->origin_in_parent.y + child->length_in_parent.y - 1;
            X = child->origin_in_parent.x;

            # pragma omp parallel for
            for (int_t x = 0; x < child->length.x; x++) {
                DARRAY(a_c, buffer, child->length.y, x) =
                    (DARRAY(a_p, 0, Y,   x/2   + X) * (refinement - t) * (refinement - x % refinement)
                   + DARRAY(a_p, B, Y,   x/2   + X) * t                * (refinement - x % refinement)
                   + DARRAY(a_p, 0, Y,   x/2+1 + X) * (refinement - t) * (x % refinement)
                   + DARRAY(a_p, B, Y,   x/2+1 + X) * t                * (x % refinement)
                   + DARRAY(a_p, 0, Y+1, x/2   + X) * (refinement - t) * (refinement - x % refinement)
                   + DARRAY(a_p, B, Y+1, x/2   + X) * t                * (refinement - x % refinement)
                   + DARRAY(a_p, 0, Y+1, x/2+1 + X) * (refinement - t) * (x % refinement)
                   + DARRAY(a_p, B, Y+1, x/2+1 + X) * t                * (x % refinement)
                   ) / (refinement * refinement * refinement);
            }

            break;

        case WEST:
            Y = child->origin_in_parent.y;
            X = child->origin_in_parent.x;

            # pragma omp parallel for
            for (int_t y = 0; y < child->length.y; y++) {
                DARRAY(a_c, buffer, y, -1) =
                    (DARRAY(a_p, 0, y/2   + Y, X)   * (refinement - t) * (refinement - y % refinement)
                   + DARRAY(a_p, B, y/2   + Y, X)   * t                * (refinement - y % refinement)
                   + DARRAY(a_p, 0, y/2+1 + Y, X)   * (refinement - t) * (y % refinement)
                   + DARRAY(a_p, B, y/2+1 + Y, X)   * t                * (y % refinement)
                   + DARRAY(a_p, 0, y/2   + Y, X-1) * (refinement - t) * (refinement - y % refinement)
                   + DARRAY(a_p, B, y/2   + Y, X-1) * t                * (refinement - y % refinement)
                   + DARRAY(a_p, 0, y/2+1 + Y, X-1) * (refinement - t) * (y % refinement)
                   + DARRAY(a_p, B, y/2+1 + Y, X-1) * t                * (y % refinement)
                   ) / (refinement * refinement * refinement);
            }

            break;

        case EAST:
            Y = child->origin_in_parent.y;
            X = child->origin_in_parent.x + child->length_in_parent.x - 1;

            # pragma omp parallel for
            for (int_t y = 0; y < child->length.y; y++) {
                DARRAY(a_c, buffer, y, child->length.x) =
                    (DARRAY(a_p, 0, y/2   + Y, X)   * (refinement - t) * (refinement - y % refinement)
                   + DARRAY(a_p, B, y/2   + Y, X)   * t                * (refinement - y % refinement)
                   + DARRAY(a_p, 0, y/2+1 + Y, X)   * (refinement - t) * (y % refinement)
                   + DARRAY(a_p, B, y/2+1 + Y, X)   * t                * (y % refinement)
                   + DARRAY(a_p, 0, y/2   + Y, X+1) * (refinement - t) * (refinement - y % refinement)
                   + DARRAY(a_p, B, y/2   + Y, X+1) * t                * (refinement - y % refinement)
                   + DARRAY(a_p, 0, y/2+1 + Y, X+1) * (refinement - t) * (y % refinement)
                   + DARRAY(a_p, B, y/2+1 + Y, X+1) * t                * (y % refinement)
                   ) / (refinement * refinement * refinement);
            }

            break;

        default:
            break;
    }
}


// The solution in the child grid is averaged down to the resolution of the
// parent grid
void downsample(grid_t *parent, grid_t *child) {
    real_t start = omp_get_wtime();
    
    # pragma omp parallel
    {
        # pragma omp for
        for (int_t y = 0; y < child->length_in_parent.y; y++) {
            for (int_t x = 0; x < child->length_in_parent.x; x++) {
                DARRAY(parent->n, 0, y + child->origin_in_parent.y,
                                     x + child->origin_in_parent.x) =
                    DARRAY(child->n, 0, y*2, x*2);
            }
        }

        # pragma omp for
        for (int_t y = 0; y < child->length_in_parent.y; y++) {
            for (int_t x = 0; x < child->length_in_parent.x; x++) {
                DARRAY(parent->u, 0, y + child->origin_in_parent.y,
                                     x + child->origin_in_parent.x) =
                    DARRAY(child->u, 0, y*2, x*2);
            }
        }

        # pragma omp for
        for (int_t y = 0; y < child->length_in_parent.y; y++) {
            for (int_t x = 0; x < child->length_in_parent.x; x++) {
                DARRAY(parent->v, 0, y + child->origin_in_parent.y,
                                     x + child->origin_in_parent.x) =
                    DARRAY(child->v, 0, y*2, x*2);
            }
        }
    }

    rank_time[downsample_index] += omp_get_wtime() - start;
}


// height_threshold: Fraction of wave height needed for a cell to be flagged.
//                  Number between 0 and 1.
// amount_threshold: Fraction of cells that should be flagged in an area for
//                  refinement to take place. Number between 0 and 1.
void regrid(grid_t *grid, real_t normal_height, real_t wave_height,
            real_t height_threshold, real_t amount_threshold)
{
    real_t start = omp_get_wtime();

    // Remove old children
    // TODO: Maybe keep the children that are identical to new grids
    //       Maybe even copy over the values from the overlapping children
    //       before they are deleted

    // for (int_t i = 0; i < grid->n_children; i++) {
    //     free_grid(grid->children[i]);
    // }

    // grid->children = realloc(grid->children, 0);
    // grid->n_children = 0;

    point_t origin = {.y = 0, .x = 0};
    rectangle_t *root = new_rectangle(origin, grid->length);
    rectangle_t **accepted = malloc(0);
    int_t n_accepted = 0;
    bool **flagged = malloc(root->length.y * sizeof(bool *));

    for (int_t y = 0; y < root->length.y; y++) {
        flagged[y] = malloc(root->length.x * sizeof(bool));
    }

    // Flag cells that require refinement.
    // a) Find elevation of sea surface relative to sea level
    # pragma omp parallel for
    for (int_t y = 0; y < root->length.y; y++) {
        for (int_t x = 0; x < root->length.x; x++) {
            // Height
            if (fabs(DARRAY(grid->n, 0, y, x) - normal_height) >
                wave_height * height_threshold * grid->total_refinement)

            // Rate of change over time
            // if (fabs(DARRAY(grid->n, 1, y, x) - DARRAY(grid->n, 0, y, x)) >
            //     1e-11 * grid->total_refinement)
            {
                flagged[y][x] = true;
            } else {
                flagged[y][x] = false;
            }
        }
    }

    // b) Alternative: Find gradient

    evaluate_rectangle(root, grid, flagged, amount_threshold, accepted,
                       &n_accepted);

    add_child_grids(grid, accepted, n_accepted);

    // Cleanup:
    // root might be free'd inside evaluate_rectangle() and can no longer be
    // used.
    for (int_t y = 0; y < grid->length.y; y++) {
        free(flagged[y]);
    }

    free(flagged);

    rank_time[regrid_index] += omp_get_wtime() - start;
}


void add_child_grids(grid_t *grid, rectangle_t **accepted, int_t n_accepted) {
    // If we always sort the rectangles and let them become the new child grids,
    // then the old child grids will always be sorted.

    // Sort the rectangles by size from largest to smallest
    qsort(accepted, n_accepted, sizeof(rectangle_t *), compare_rectangles);

    // Attempt to pair the largest (remaining) old grid with the
    // largest (remaining) new rectangle.
    for (int_t i = 0; i < n_accepted || i < grid->n_children; i++) {
        if (i < n_accepted && i < grid->n_children) {
            // Pair grid and rectangle.
            // Reuse grid to make new grid from rectangle.

            // TODO: Keep children of children in an old array in case we want
            //       to allocate children of children in the future?

            // add_child_grid(grid, accepted[i]->origin, accepted[i]->length,
            //                grid->children[i]);

            update_child_grid(grid->children[i], grid, accepted[i]->origin,
                              accepted[i]->length, 2);

            free_rectangle(accepted[i]);
        } else if (i < n_accepted) {
            // More rectangles than grids. Make new grids.
            add_child_grid(grid, accepted[i]->origin, accepted[i]->length);
            free_rectangle(accepted[i]);
        } else if (i < grid->n_children) {
            // More grids than rectangles. Free spare grids.
            // remove_child_grid(grid->children[i], grid);
            free_grid(grid->children[i]);
        }
    }

    free(accepted);

    // If more grids than rectangles, reduce the number of children.
    // The children themselves have already been freed.
    int_t n_spare_grids = grid->n_children - n_accepted;

    if (n_spare_grids > 0) {
        grid->children =
            realloc(grid->children,
                    (grid->n_children - n_spare_grids) * sizeof(grid_t *));

        grid->n_children -= n_spare_grids;
    }
}


// wave_height: Max height for wave relative to normal height
void initialize(grid_t *grid, length_t length, real_t normal_height,
                real_t wave_height, wave_t wave)
{
    /* Stupid but simple algorithm, only runs at initialization anyway */

    /* Iterate across the entire domain */
    for (int_t y = 0; y < length.y; y++) {
        for (int_t x = 0; x < length.x; x++) {
            /* Check if this point is in my local domain */
            if (y >= grid->origin.y
             && x >= grid->origin.x
             && y <  grid->origin.y + grid->length.y
             && x <  grid->origin.x + grid->length.x
            )
            {
                // Set normal level
                DARRAY(grid->n, 0, y - grid->origin.y, x - grid->origin.x) =
                    normal_height;

                if (wave == edge) {
                    // Wave from edge
                    if (y < length.y / 4) {
                        DARRAY(grid->n, 0, y - grid->origin.y, x - grid->origin.x) +=
                        wave_height * (cos(y * acos(-1) / (length.y / 4)) + 1) / 2;
                    }
                } else if (wave == constant_length_edge) {
                    // Like wave from edge, but with a constant length instead of a length which depends on the length of the domain
                    if (y < 128) {
                        DARRAY(grid->n, 0, y - grid->origin.y, x - grid->origin.x) +=
                        wave_height * (cos(y * acos(-1) / (length.y / 4)) + 1) / 2;
                    }
                } else if (wave == middle) {
                    // Wave from middle 
                    if (y > length.y / 4 && y < 3 * length.y / 4) {
                        DARRAY(grid->n, 0, y - grid->origin.y, x - grid->origin.x) +=
                        wave_height * (cos(y * acos(-1) / (length.y / 4)) + 1) / 2;
                    }
                } else if (wave == center) {
                    // Offset level if coordinates are near center of domain
                    if (y > length.y / 4 && y < 3 * length.y / 4 &&
                        x > length.x / 4 && x < 3 * length.x / 4)
                    {
                        DARRAY(grid->n, 0, y - grid->origin.y, x - grid->origin.x) +=
                            normal_height * exp (
                            - 4 * pow(x - length.x / 2, 2.0) / (real_t)(length.x)
                            - 4 * pow(y - length.y / 2, 2.0) / (real_t)(length.y)
                        );
                    }
                }
            }
        }
    }

    mirror_boundary(grid, grid->n, 0);
}


// r: Refinement
void initialize_child_grids(grid_t *parent, grid_t *child, int_t r) {
    real_t start = omp_get_wtime();
    
    // Set everything up to and including border (offset 0)
    initialize_child_grid(child, child->n, parent->n, r);
    initialize_child_grid(child, child->u, parent->u, r);
    initialize_child_grid(child, child->v, parent->v, r);
    initialize_child_grid(child, child->pnu, parent->pnu, r);
    initialize_child_grid(child, child->pnv, parent->pnv, r);

    // Set outside border (offset -1)
    set_grid_borders(child, parent->n,   child->n,   0, r, 0);
    set_grid_borders(child, parent->u,   child->u,   0, r, 0);
    set_grid_borders(child, parent->v,   child->v,   0, r, 0);
    set_grid_borders(child, parent->pnu, child->pnu, 0, r, 0);
    set_grid_borders(child, parent->pnv, child->pnv, 0, r, 0);

    mirror_boundary(child, child->n, 0);
    reverse_boundary(child, child->u, 0);
    reverse_boundary(child, child->v, 0);

    rank_time[initialize_child_grids_index] += omp_get_wtime() - start;
}


// a_c: Child array
// a_p: Parent array
// r: Refinement
void initialize_child_grid(grid_t *child, dist_array_t *a_c, dist_array_t *a_p,
                           int_t r) {
    int_t
        Y = child->origin_in_parent.y,
        X = child->origin_in_parent.x;

    # pragma omp parallel for
    for (int_t y = 0; y < child->length.y; y++) {
        for (int_t x = 0; x < child->length.x; x++) {
            DARRAY(a_c, 0, y, x) =
                (DARRAY(a_p, 0, y/r   + Y, x/r   + X) * (r - y % r) * (r - x % r)
               + DARRAY(a_p, 0, y/r   + Y, x/r+1 + X) * (r - y % r) * (x % r)
               + DARRAY(a_p, 0, y/r+1 + Y, x/r   + X) * (y % r)     * (r - x % r)
               + DARRAY(a_p, 0, y/r+1 + Y, x/r+1 + X) * (y % r)     * (x % r)
                ) / (r * r); // In y and x direction
        }
    }
}


/* Dump buffer 0 to file */
void write_output(grid_t *grid, const char *filename) {
    for (int r=0; r<size; r++) {
        if (rank == r) {
            FILE *out = fopen(filename, (rank == 0 && grid->level == 0) ? "w" : "a");

            for (int_t i = 0; i < grid->length.y; i++) {
                for (int_t j = 0; j < grid->length.x; j++) {
                    if (isnan(DARRAY(grid->n, 0, i, j))) {
                        // Comment which will be ignored by gnuplot
                        fprintf(out, "# ");
                    }

                    fprintf(out, "%f %f %e Level: %" PRId64 "\n",
                            i / pow(2, grid->level) + grid->origin.y,
                            j / pow(2, grid->level) + grid->origin.x,
                            DARRAY(grid->n, 0, i, j), grid->level);
                }
            }

            fclose(out);

            for (int_t i = 0; i < grid->n_children; i++) {
                write_output(grid->children[i], filename);
            }
        }

        if (grid->level == 0) { // Root
            MPI_Barrier(cart);
        }
    }
}


void free_grid(grid_t *grid) {
    free_dist_array(grid->n);
    free_dist_array(grid->pn);
    free_dist_array(grid->pnu);
    free_dist_array(grid->pnv);
    free_dist_array(grid->pnuv);
    free_dist_array(grid->du);
    free_dist_array(grid->dv);
    free_dist_array(grid->u);
    free_dist_array(grid->v);

    for (int_t i = 0; i < grid->n_children; i++) {
        free_grid(grid->children[i]);
    }

    free(grid->children);
    free(grid);
}
