#include <omp.h>

#include "rectangle.h"


extern real_t* rank_time;


rectangle_t * new_rectangle(point_t origin, length_t length) {
    rectangle_t *r = malloc(sizeof(rectangle_t));

    r->origin = origin;
    r->length = length;

    // Make it possible to free all rectangles the same way, regardless if we
    // need to realloc more space later or not
    r->signature.y = malloc(0);
    r->signature.x = malloc(0);

    r->laplacian.y = malloc(0);
    r->laplacian.x = malloc(0);

    return r;
}


void accept_rectangle(rectangle_t *r, rectangle_t **accepted,
                      int_t *n_accepted)
{
    accepted =
        realloc(accepted,
                (*n_accepted + 1) * sizeof(rectangle_t *));

    (*n_accepted)++;
    accepted[*n_accepted - 1] = r;
}


// Sort the rectangles from largest to smallest
int compare_rectangles(const void *p1, const void *p2) {
    // qsort() calls our compare function with a void pointer to each element in
    // our array.
    // Our array contains rectangle_t pointers.
    // That means qsort() calls our compare function with a void double pointer.
    // We want a rectangle_t pointer.
    // To get that, we first cast the void double pointer to a rectangle_t
    // double pointer.
    // Then we dereference the rectangle_t double pointer to get a rectangle_t
    // pointer.
    rectangle_t *r1 = *(rectangle_t **)p1;
    rectangle_t *r2 = *(rectangle_t **)p2;

    int_t border = 1; // TODO: Not hardcode. See if qsort_r() can help.

    int_t size_1 = (r1->length.y + 2 * border) * (r1->length.x + 2 * border);
    int_t size_2 = (r2->length.y + 2 * border) * (r2->length.x + 2 * border);

    return size_2 - size_1; // If r1 is larger, it will go before r2
}


// Recursively cluster the flagged cells into rectangular patches
void evaluate_rectangle(rectangle_t *r, grid_t *grid, bool **flagged,
                        real_t threshold, rectangle_t **accepted,
                        int_t *n_accepted)
{
    int_t n_flagged = 0;

    # pragma omp parallel for reduction(+: n_flagged)
    for (int_t y = r->origin.y; y < r->origin.y + r->length.y; y++) {
        for (int_t x = r->origin.x; x < r->origin.x + r->length.x; x++) {
            if (flagged[y][x]) {
                n_flagged++;
            }
        }
    }

    if (n_flagged == 0) {
        // Empty rectangle, does not need to be refined
        free_rectangle(r);
        return;
    }

    real_t efficiency = (real_t)n_flagged / (r->length.y * r->length.x);

    if (efficiency < threshold) {
        // Find the best place to split rectangle
        // Either a hole or inflection point

        // Opitimization: Only allocate signature and laplacian if needed
        r->signature.y = realloc(r->signature.y, r->length.y * sizeof(int_t));
        r->signature.x = realloc(r->signature.x, r->length.x * sizeof(int_t));

        // Signature is only incremented and therefore needs to be initialized
        memset(r->signature.y, 0, r->length.y * sizeof(int_t));
        memset(r->signature.x, 0, r->length.x * sizeof(int_t));

        // Waste 1 index (the first one) for easier indexing
        r->laplacian.y = malloc((r->length.y - 1) * sizeof(int_t));
        r->laplacian.x = malloc((r->length.x - 1) * sizeof(int_t));

        // Compute signatures.
        // Use 2 double for loops with different order of y and x to avoid race
        // condition.
        // OpenMP 4.5 has a convenient array reduction syntax, but at the cost
        // of overhead from malloc'ing for each thread. This approach avoids
        // that at the cost of more looping and double the if checks.
        // Donald Knuth cries in the corner at my premature optimizations.
        # pragma omp parallel for
        for (int_t y = 0; y < r->length.y; y++) {
            for (int_t x = 0; x < r->length.x; x++) {
                if (flagged[r->origin.y + y][r->origin.x + x]) {
                    // Safe because only one thread holds a given y
                    r->signature.y[y]++;
                }
            }
        }

        # pragma omp parallel for
        for (int_t x = 0; x < r->length.x; x++) {
            for (int_t y = 0; y < r->length.y; y++) {
                if (flagged[r->origin.y + y][r->origin.x + x]) {
                    // Safe because only one thread holds a given x
                    r->signature.x[x]++;
                }
            }
        }

        direction_t split_direction;
        int_t split_index;

        // First look for holes
        bool found_place_to_split =
            find_hole(r, &split_index, &split_direction);

        if (! found_place_to_split) {
            found_place_to_split =
                find_inflection_point(r, &split_index, &split_direction);
        }

        if (found_place_to_split) {
            // Split rectangle in two
            point_t origin_second;

            length_t
                length_first,
                length_second;

            if (split_direction == horizontal) {
                origin_second = (point_t){.y = split_index, .x = r->origin.x};
                length_first  =
                    (length_t){.y = split_index - r->origin.y,
                               .x = r->length.x};

                length_second =
                    (length_t){.y = r->origin.y + r->length.y - split_index,
                               .x = r->length.x};
            } else {
                origin_second = (point_t){.y = r->origin.y, .x = split_index};
                length_first  =
                    (length_t){.y = r->length.y,
                               .x = split_index - r->origin.x};

                length_second =
                    (length_t){.y = r->length.y,
                               .x = r->origin.x + r->length.x - split_index};
            }

            rectangle_t *first = new_rectangle(r->origin, length_first);
            rectangle_t *second = new_rectangle(origin_second, length_second);

            free_rectangle(r);

            evaluate_rectangle(first, grid, flagged, threshold, accepted,
                               n_accepted);

            evaluate_rectangle(second, grid, flagged, threshold, accepted,
                               n_accepted);
        } else {
            // Accept the sub-treshold rectangle because it can not be split
            // further
            // real_t start = omp_get_wtime();

            // add_child_grid(grid, r->origin, r->length);
            accept_rectangle(r, accepted, n_accepted);

            // rank_time[add_child_grid_index] += omp_get_wtime() - start;

            // free_rectangle(r);
        }
    } else {
        // Accept the rectangle because it satisfies the treshold
        // real_t start = omp_get_wtime();

        // add_child_grid(grid, r->origin, r->length);
        accept_rectangle(r, accepted, n_accepted);

        // rank_time[add_child_grid_index] += omp_get_wtime() - start;

        // free_rectangle(r);
    }
}


// Find index of hole in a given direction.
// Return true if hole found, false otherwise.
// Input:
//     signature: 1D array of signature in a given direction.
//     length: Length of signature array.
//     new_direction: The direction to set split_direction to if a hole is
//     found.
//     origin: Origin of rectangle in a given direction.
// Output
//     split_index: Index to split at.
//     split_direction: Direction to split in, either horizontal or vertical.
bool find_hole_in_direction(int_t *signature, int_t length, int_t *split_index,
                            direction_t *split_direction,
                            direction_t new_direction, int_t origin)
{
    // TODO: Run in parallel
    // # pragma omp parallel for
    for (int_t i = 1; i < length; i++) {
        if ((signature[i-1] == 0) != (signature[i] == 0)) {
            *split_index = origin + i;
            *split_direction = new_direction;
            return true;
        }
    }

    return false;
}


// Find index of hole and direction to split rectangle in
// Return true if hole found, false otherwise
// Input:
//     r: Rectangle
// Output:
//     split_index: Index to split at
//     split_direction: Direction to split in, either horizontal or vertical
bool find_hole(rectangle_t *r, int_t *split_index,
               direction_t *split_direction)
{
    // Split at transition from holes to not holes or opposite
    // Should I split at the longest hole, or at any hole?
    // I currently split at any hole

    // || (double or?) only evaluates the first expression if it is found to be
    // true, so in case a hole is found in the y direction it is not
    // overwritten if a hole is also found in the x direction.
    return find_hole_in_direction(r->signature.y, r->length.y, split_index,
                                  split_direction, horizontal, r->origin.y)
        || find_hole_in_direction(r->signature.x, r->length.x, split_index,
                                  split_direction, vertical, r->origin.x);
}


// Find index of inflection point and direction to split rectangle in
// Return true if inflection point found, false otherwise
// Input:
//     r: Rectangle
// Output:
//     split_index: Index to split at
//     split_direction: Direction to split in, either horizontal or vertical
bool find_inflection_point(rectangle_t *r, int_t *split_index,
                           direction_t *split_direction)
{
    // An inflection point is a point where the laplacian crosses zero

    // Find the laplacian of the signatures
    get_laplacian(r->length.y, r->signature.y, r->laplacian.y);
    get_laplacian(r->length.x, r->signature.x, r->laplacian.x);

    // Find the largest zero crossing
    int_t max_zero_crossing_value = 0;

    get_max_zero_crossing(&max_zero_crossing_value, split_index,
                          split_direction, r->length.y, r->laplacian.y,
                          horizontal, r->origin.y);

    get_max_zero_crossing(&max_zero_crossing_value, split_index,
                          split_direction, r->length.x, r->laplacian.x,
                          vertical, r->origin.x);

    if (max_zero_crossing_value > 0) {
        return true;
    }

    return false;
}


// Find laplacian of signatures
// Input:
//     length: Length of signature and laplacian arrays
//     signature: 1D array of signatures
// Output:
//     laplacian: 1D array of laplacian of signatures
void get_laplacian(int_t length, int_t *signature, int_t *laplacian) {
    # pragma omp parallel for
    for (int_t i = 1; i < length - 1; i++) {
        // 1D laplacian using FDM = x[i-1] - 2*x[i] + x[i+1]
        laplacian[i] = signature[i-1] - 2*signature[i] + signature[i+1];
    }
}


// Input and output:
//     max_zero_crossing_value: Value of the largest zero crossing
// Output:
//     max_zero_crossing_index: Index of the largest zero crossing
//     split_direction: Which direction to split the rectangle
// Input:
//     length: Length of laplacian array
//     laplacian: 1D array of the laplacian of the signature
//     origin: Origin of rectangle in a given direction
void get_max_zero_crossing(int_t *max_zero_crossing_value, int_t *split_index,
                           direction_t *split_direction, int_t length,
                           int_t *laplacian, direction_t new_direction,
                           int_t origin)
{
    // TODO: Make it safe to run this in parallel
    for (int_t i = 2; i < length - 1; i++) {
        if (is_zero_crossing(laplacian[i-1], laplacian[i])) {
            int_t zero_crossing_value =
                llabs(laplacian[i-1] - laplacian[i]);

            if (zero_crossing_value > *max_zero_crossing_value) {
                *max_zero_crossing_value = zero_crossing_value;
                *split_index = origin + i;
                *split_direction = new_direction;
            }
        }
    }
}


bool is_zero_crossing(int_t a, int_t b) {
    // Requirements:
    // Should either go from positive to negative number,
    // or from positive to zero,
    // or from negative to zero,
    // but not from zero to zero.

    return
    // Different sign
    ((a < 0) != (b < 0))
    ||
    // Exactly one of them are zero
    ((a == 0) ^ (b == 0));
}


void free_rectangle(rectangle_t *r) {
    free(r->signature.y);
    free(r->signature.x);

    free(r->laplacian.y);
    free(r->laplacian.x);

    free(r);
}
