#include <inttypes.h>
#include <unistd.h>
#include <omp.h>

#include "swe.h"
#include "dist_array.h"
#include "grid.h"


dim_type_t dim_type = manual;


// Profiling
const int_t n_profile_functions = 17;


// Array of accumulated time measurements for different functions. Each rank has
// its own array.
real_t* rank_time;


const char* profile_function_names[] =
    {"add child grid", "allocate child grid", "boundary", "downsample",
     "grid_border", "intialize child grids", "integrate time step", "malloc",
     "memset_malloc", "memset_realloc", "mpi type", "rank_border",
     "update local state", "update time step", "realloc", "regrid", "wait"};


/* Problem parameters */
int_t
    Y = 80, // Default width
    X = 80, // Default length

    max_refinement = 2,

    // Number of timesteps between each regridding.
    // TODO: Set this automatically
    regrid_interval = 5,

    // The number of points for a given rank for all iterations
    n_rank_points = 0,
    n_rank_child_points = 0,

    // The total number of points for all ranks for all iterations
    n_total_points = 0,
    n_total_child_points = 0,

    // The number of points for a given rank for all iterations, measured at
    // the previous iteration for root.
    // Used to find the number of points for a given rank in a given iteration
    // for root as the difference between the current number of points and the
    // value measured at the previous iteration for root.
    // We measure between iterations for root because root runs blocking
    // communication every iteration during border exchange.
    n_rank_points_last_root_iter = 0,

    // The sum over all iterations for root of the the maximum number of point
    // for a single rank in a given iteration for root.

    // This indicates the real cost of load imbalance. If one rank has a
    // heavier load than the other ranks for a given iteration for root, the
    // other ranks will have to wait for the first rank. If a different rank has
    // the heaviest load in the next iteration for root, the waiting will not
    // cancel out, but rather add up.
    n_total_load_points = 0,

    max_iter = 40001,
    snapshot = 200,
    global[2];


bool
    debug = false,
    measure_load_balance = false,
    output = false,
    border_exchange_barrier = true;


real_t
    normal_height = 1,
    wave_height = 1,
    height_threshold = 0.1,

    // Higher threshold ->
    //     Smaller rectangles -> Less unnecessary work
    //     More rectangles    -> Larger boundary condition overhead
    amount_threshold = 0.9,

    dy = 1,
    dx = 1,
    dt = 1.25e-2;


// Origin of start wave
wave_t wave = edge;


/* Physical parameters (and shorthand notation) */
const real_t
    gravity = 9.81,
    density = 997.0;


#define G (gravity)
#define P (density)


/* Function prototypes */
void run(grid_t * grid, int_t T);
void parse_arguments (int argc, char **argv);
void update_local_state(grid_t *grid);
void set_boundary(grid_t *grid, dist_array_t *m, int_t timestep, int_t sign);
void integrate_time_step(grid_t *grid);
void update_time_step(grid_t *grid);
void print_output(real_t total_rank_time);
void update_load_balance_measure();


/* Program entry point */
int main (int argc, char **argv) {
    parse_arguments (argc, argv);

    /* Initialize Cartesian topology that is used to partition all arrays */
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if (dim_type == manual) {
        dims[0] = 1;
        dims[1] = size;
    } else {
        MPI_Dims_create(size, 2, dims);
    }

    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, (int[2]){0,0}, 0, &cart);
    MPI_Comm_rank(cart, &rank);
    MPI_Cart_coords(cart, rank, 2, coords);
    MPI_Cart_shift(cart, 0, 1, &north, &south);
    MPI_Cart_shift(cart, 1, 1, &west, &east);

    // Profiling
    rank_time = calloc(n_profile_functions, sizeof(real_t));
    MPI_Barrier(cart);
    real_t start = MPI_Wtime();
    
    point_t origin = {.y = (global[0] / dims[0]) * coords[0]
                         + (global[0] % dims[0]) * coords[0],
                      .x = (global[1] / dims[1]) * coords[1]
                         + (global[1] % dims[1]) * coords[1]};

    length_t length = {.y = global[0] / dims[0]
                          + ((coords[0] < (global[0] % dims[0])) ? 1 : 0),
                       .x = global[1] / dims[1]
                          + ((coords[1] < (global[1] % dims[1])) ? 1 : 0)};

    length_t global_length = {.y = global[0], .x = global[1]};
    dt *= dy; // Keep CFL valid
    grid_t *root_grid = new_root_grid(origin, length, dy, dx, dt);
    initialize(root_grid, global_length, normal_height, wave_height, wave);

    if (debug) {
        if (rank == 0) {
            printf("Connect your debugger using the following PID(s):\n");
        }

        MPI_Barrier(cart);
        printf("PID: %d\n", getpid()); // Used for connecting a debugger
        MPI_Barrier(cart);

        if (rank == 0) {
            printf("Press any key to continue...\n");
        }

        getchar(); // Pause program until keypress
    }

    run(root_grid, max_iter);
    free_grid(root_grid);
    real_t total_rank_time = MPI_Wtime() - start;
    print_output(total_rank_time);
    MPI_Finalize();
}


// Recursively run all the timesteps for a grid and its children
// T: Number of timesteps
void run(grid_t * grid, int_t T) {
    // TODO: Do this in some much smarter and less hacky way.
    //       At least look into if it's enough to only do it at
    //       initialization.
    DARRAY(grid->u, 1, grid->length.y, grid->length.x) = 0;
    DARRAY(grid->v, 1, grid->length.y, grid->length.x) = 0;

    DARRAY(grid->u, 1, -1, grid->length.x) = 0;
    DARRAY(grid->v, 1, -1, grid->length.x) = 0;

    DARRAY(grid->u, 1, grid->length.y, -1) = 0;
    DARRAY(grid->v, 1, grid->length.y, -1) = 0;

    for (int_t iter = 0; iter < T; iter++) {
        int_t points = grid->length.y * grid->length.x;
        n_rank_points += points;

        if (grid->level >= 1) {
            n_rank_child_points += points;
        }

        if (output && grid->level == 0 && iter % snapshot == 0) {
            char filename[256];
            sprintf(filename, "data/state_%.5ld.txt", iter / snapshot);
            write_output(grid, filename);

            if (rank == 0) {
                printf("Iter %ld\n", iter);
            }
        }

        if (grid->parent != NULL) { // Root has no parent
            set_grid_borders(grid, grid->parent->n,   grid->n,   iter, 2, 0);
            set_grid_borders(grid, grid->parent->u,   grid->u,   iter, 2, 0);
            set_grid_borders(grid, grid->parent->v,   grid->v,   iter, 2, 0);
            set_grid_borders(grid, grid->parent->pnu, grid->pnu, iter, 2, 0);
            set_grid_borders(grid, grid->parent->pnv, grid->pnv, iter, 2, 0);
        }

        if (grid->level == 0) {
            border_exchange(grid->u, 0);
            border_exchange(grid->v, 0);
        }

        update_local_state(grid);

        // TODO: Replace requirement with whether grid has neighbor(s)
        if (grid->level == 0) { // Root
            border_exchange(grid->du, 0);
            border_exchange(grid->dv, 0);
            border_exchange(grid->pn, 0);
            border_exchange(grid->pnu, 0);
            border_exchange(grid->pnv, 0);
            border_exchange(grid->pnuv, 0);
        }

        mirror_boundary(grid, grid->pn, 0);
        reverse_boundary(grid, grid->pnu, 0);
        reverse_boundary(grid, grid->pnv, 0);

        integrate_time_step(grid);

        if (grid->parent != NULL) {
            set_grid_borders(grid, grid->parent->pn,  grid->pn,  iter+1, 2, 2);
            set_grid_borders(grid, grid->parent->pnu, grid->pnu, iter+1, 2, 2);
            set_grid_borders(grid, grid->parent->pnv, grid->pnv, iter+1, 2, 2);
        }

        mirror_boundary(grid, grid->pn, 2);
        reverse_boundary(grid, grid->pnu, 2);
        reverse_boundary(grid, grid->pnv, 2);

        if (grid->level == 0) {
            border_exchange(grid->pn, 2);
            border_exchange(grid->pnu, 2);
            border_exchange(grid->pnv, 2);
        }

        update_time_step(grid);

        // Regridding
        if (grid->regrid_ticks % regrid_interval == 0 &&
            grid->total_refinement * 2 <= max_refinement)
        {
            regrid(grid, normal_height, wave_height, height_threshold,
                   amount_threshold);
        }

        grid->regrid_ticks = (grid->regrid_ticks + 1) % regrid_interval;

        for (int_t i = 0; i < grid->n_children; i++) {
            run(grid->children[i], 2);
            downsample(grid, grid->children[i]);

            // TODO: Adjust the coarse cell values adjacent to fine grids to
            // maintaint conservation of mass (and momentum in regions where
            // the source terms vanish)
        }

        swap_buffers(grid->pn);
        swap_buffers(grid->pnu);
        swap_buffers(grid->pnv);
        swap_buffers(grid->n);
        swap_buffers(grid->u);
        swap_buffers(grid->v);

        // Update how many points have been handled

        // Call this every iteration for root, since root runs blocking border
        // exchange every iteration during border exchange.
        // Only call if we are measuring load balance, to avoid overhead if not
        if (measure_load_balance && grid->level == 0) {
            update_load_balance_measure();
        }
    }
}


void parse_arguments(int argc, char **argv) {
    for (int i = 1; i < argc; i++) {
        if (strncmp(argv[i], "-Y", 2) == 0) {
            Y = strtoll(argv[i + 1], NULL, 10);
        } else if (strncmp(argv[i], "-X", 2) == 0) {
            X = strtoll(argv[i + 1], NULL, 10);
        } else if (strncmp(argv[i], "--debug", 7) == 0) {
            debug = true;
        } else if (strncmp(argv[i], "--output", 8) == 0) {
            output = true;
        } else if (strncmp(argv[i], "--normal_height", 15) == 0) {
            normal_height = strtod(argv[i+1], NULL);
        } else if (strncmp(argv[i], "--wave_height", 13) == 0) {
            wave_height = strtod(argv[i+1], NULL);
        } else if (strncmp(argv[i], "--height_threshold", 18) == 0) {
            height_threshold = strtod(argv[i+1], NULL);
        } else if (strncmp(argv[i], "--amount_threshold", 18) == 0) {
            amount_threshold = strtod(argv[i+1], NULL);
        } else if (strncmp(argv[i], "--max_refinement", 16) == 0) {
            max_refinement = strtoll(argv[i+1], NULL, 10);
        } else if (strncmp(argv[i], "--dy", 4) == 0) {
            dy = strtod(argv[i+1], NULL);
        } else if (strncmp(argv[i], "--dx", 4) == 0) {
            dx = strtod(argv[i+1], NULL);
        } else if (strncmp(argv[i], "--regrid_interval", 17) == 0) {
            regrid_interval = strtoll(argv[i+1], NULL, 10);
        } else if (strncmp(argv[i], "--measure_load_balance", 22) == 0) {
            measure_load_balance = true;
        } else if (strncmp(argv[i], "--max_iter", 10) == 0) {
            max_iter = strtoll(argv[i+1], NULL, 10);
        } else if (strncmp(argv[i], "--wave", 6) == 0) {
            if (strncmp(argv[i+1], "edge", 4) == 0) {
                wave = edge;
            } else if (strncmp(argv[i+1], "center", 6) == 0) {
                wave = center;
            } else if (strncmp(argv[i+1], "middle", 6) == 0) {
                wave = middle;
            } else if (strncmp(argv[i+1], "constant_length_edge", 20) == 0) {
                wave = constant_length_edge;
            }
        } else if (strncmp(argv[i], "--dim_type", 10) == 0) {
            if (strncmp(argv[i+1], "manual", 6) == 0) {
                dim_type = manual;
            } else if (strncmp(argv[i+1], "automatic", 9) == 0) {
                dim_type = automatic;
            }
        } else if (strncmp(argv[i], "--border_exchange_barrier", 25) == 0) {
            if (strncmp(argv[i+1], "false", 5) == 0) {
                border_exchange_barrier = false;
            }
        }
    }

    global[0] = Y;
    global[1] = X;
}


void update_local_state(grid_t *grid) {
    real_t start = omp_get_wtime();
    
    # pragma omp parallel
    {
        # pragma omp for
        // Integration requires the following to exist in the domain exterior.
        for (int_t y = -grid->border; y < grid->length.y + grid->border; y++)
            for (int_t x = -grid->border; x < grid->length.x + grid->border; x++)
                DARRAY(grid->pn, 0, y, x) = P * DARRAY(grid->n, 0, y, x);

        # pragma omp for
        // The parts that are written into halo regions in these 3 will shortly
        // be overwritten by border exchanges, but this is simpler code for now.
        for (int_t y = -grid->border; y < grid->length.y + grid->border; y++)
            for (int_t x = -grid->border; x < grid->length.x + grid->border; x++)
                DARRAY(grid->pnuv, 0, y, x) =
                    P * DARRAY(grid->n, 0, y, x) * DARRAY(grid->u, 0, y, x)
                      * DARRAY(grid->v, 0, y, x);

        # pragma omp for
        for (int_t y = -grid->border; y < grid->length.y + grid->border; y++)
            for (int_t x = -grid->border; x < grid->length.x + grid->border; x++)
                DARRAY(grid->du, 0, y, x) =
                    P * DARRAY(grid->n, 0, y, x) * DARRAY(grid->u, 0, y, x)
                      * DARRAY(grid->u, 0, y, x)
                  + 0.5 * P * G * DARRAY(grid->n, 0, y, x)
                                * DARRAY(grid->n, 0, y, x);

        # pragma omp for
        for (int_t y = -grid->border; y<grid->length.y + grid->border; y++)
            for (int_t x = -grid->border; x < grid->length.x + grid->border; x++)
                DARRAY(grid->dv, 0, y, x) =
                    P * DARRAY(grid->n, 0, y, x) * DARRAY(grid->v, 0, y, x)
                      * DARRAY(grid->v, 0, y, x)
                  + 0.5 * P * G * DARRAY(grid->n, 0, y, x)
                                * DARRAY(grid->n, 0, y, x);
    }

    rank_time[update_local_state_index] += omp_get_wtime() - start;
}


void reverse_boundary(grid_t *grid, dist_array_t *m, int_t timestep) {
    set_boundary(grid, m, timestep, -1);
}


void mirror_boundary(grid_t *grid, dist_array_t *m, int_t timestep) {
    set_boundary(grid, m, timestep, 1);
}


// Helper method for reverse_boundary() and mirror_boundary().
// sign: Either 1 (positive) or -1 (negative)
void set_boundary(grid_t *grid, dist_array_t *m, int_t timestep, int_t sign) {
    real_t start = omp_get_wtime();
    
    bool
        is_north = grid->origin.y == 0,
        is_south = grid->global_end.y == global[0] - 1,
        is_west = grid->origin.x == 0,
        is_east = grid->global_end.x == global[1] - 1;

    if (is_north)
        for (int_t r = 1; r <= grid->border; r++)
            for (int_t x = 0; x < grid->length.x; x++)
                DARRAY(m, timestep, -r, x) = sign * DARRAY(m, timestep, r, x);

    if (is_south)
        for (int_t r = 1; r <= grid->border; r++)
            for (int_t x = 0; x < grid->length.x; x++)
                DARRAY(m, timestep, grid->length.y - 1 + r, x) =
                    sign * DARRAY(m, timestep, grid->length.y - 1 - r, x);

    if (is_west)
        for (int_t r = 1; r <= grid->border; r++)
            for (int_t y = 0; y < grid->length.y; y++)
                DARRAY(m, timestep, y, -r) = sign * DARRAY(m, timestep, y, r);

    if (is_east)
        for (int_t r = 1; r <= grid->border; r++)
            for (int_t y = 0; y < grid->length.y; y++)
                DARRAY(m, timestep, y, grid->length.x - 1 + r) =
                    sign * DARRAY(m, timestep, y, grid->length.x - 1 - r);

    // Corners
    if (is_north && is_west)
        for (int_t y = 1; y <= grid->border; y++)
            for (int_t x = 1; x <= grid->border; x++)
                DARRAY(m, timestep, -y, -x) = sign * DARRAY(m, timestep, y, x);

    if (is_north && is_east)
        for (int_t y = 1; y <= grid->border; y++)
            for (int_t x = 1; x <= grid->border; x++)
                DARRAY(m, timestep, -y, grid->length.x - 1 + x) =
                    sign * DARRAY(m, timestep, y, grid->length.x - 1 - x);

    if (is_south && is_west)
        for (int_t y = 1; y <= grid->border; y++)
            for (int_t x = 1; x <= grid->border; x++)
                DARRAY(m, timestep, grid->length.y - 1 + y, -x) =
                    sign * DARRAY(m, timestep, grid->length.y - 1 - y, x);

    if (is_south && is_east)
        for (int_t y = 1; y <= grid->border; y++)
            for (int_t x = 1; x <= grid->border; x++)
                DARRAY(m, timestep, grid->length.y - 1 + y, grid->length.x - 1 + x) =
                    sign * DARRAY(m, timestep, grid->length.y - 1 - y, grid->length.x - 1 - x);

    rank_time[boundary_index] += omp_get_wtime() - start;
}


void integrate_time_step (grid_t *grid) {
    real_t start = omp_get_wtime();
    
    // Predictor step
    # pragma omp parallel
    {
        // Velocity x predictor
        # pragma omp for
        for ( int_t y=0; y<grid->length.y; y++ )
            for ( int_t x=0; x<grid->length.x; x++ )
                DARRAY(grid->pnu,1,y,x) = DARRAY(grid->pnu,0,y,x) - grid->dt * (
                    (DARRAY(grid->du,0,y,x+1)   - DARRAY(grid->du,0,y,x))   / grid->dx
                  + (DARRAY(grid->pnuv,0,y+1,x) - DARRAY(grid->pnuv,0,y,x)) / grid->dy
                );

        // Velocity y predictor
        # pragma omp for
        for ( int_t y=0; y<grid->length.y; y++ )
            for ( int_t x=0; x<grid->length.x; x++ )
                DARRAY(grid->pnv,1,y,x) = DARRAY(grid->pnv,0,y,x) - grid->dt * (
                    (DARRAY(grid->dv,0,y+1,x)   - DARRAY(grid->dv,0,y,x))   / grid->dy
                  + (DARRAY(grid->pnuv,0,y,x+1) - DARRAY(grid->pnuv,0,y,x)) / grid->dx
                );

        // Time predictor
        # pragma omp for
        for ( int_t y=0; y<grid->length.y; y++ )
            for ( int_t x=0; x<grid->length.x; x++ )
                DARRAY(grid->pn,1,y,x) = DARRAY(grid->pn,0,y,x) - grid->dt * (
                    (DARRAY(grid->pnu,0,y,x+1) - DARRAY(grid->pnu,0,y,x)) / grid->dx
                  + (DARRAY(grid->pnv,0,y+1,x) - DARRAY(grid->pnv,0,y,x)) / grid->dy
                );
    }

    rank_time[integrate_time_step_index] += omp_get_wtime() - start;

    mirror_boundary(grid, grid->pn, 1);
    reverse_boundary(grid, grid->pnu, 1);
    reverse_boundary(grid, grid->pnv, 1);

    if (grid->level == 0) { // Root
        border_exchange(grid->pn, 1);
        border_exchange(grid->pnu, 1);
        border_exchange(grid->pnv, 1);
    }

    start = omp_get_wtime();

    // Corrector step
    # pragma omp parallel
    {
        // Velocity x corrector
        # pragma omp for
        for ( int_t y=0; y<grid->length.y; y++ )
            for ( int_t x=0; x<grid->length.x; x++ )
                DARRAY(grid->pnu,2,y,x) =
                    (DARRAY(grid->pnu,0,y,x) + DARRAY(grid->pnu,1,y,x)) / 2
                  - (grid->dt / 2) * (
                    (DARRAY(grid->du,0,y,x)   - DARRAY(grid->du,0,y,x-1))   / grid->dx
                  + (DARRAY(grid->pnuv,0,y,x) - DARRAY(grid->pnuv,0,y-1,x)) / grid->dy
                );

        // Velocity y corrector
        # pragma omp for
        for ( int_t y=0; y<grid->length.y; y++ )
            for ( int_t x=0; x<grid->length.x; x++ )
                DARRAY(grid->pnv,2,y,x) =
                    (DARRAY(grid->pnv,0,y,x) + DARRAY(grid->pnv,1,y,x)) / 2
                  - (grid->dt / 2) * (
                    (DARRAY(grid->dv,0,y,x)   - DARRAY(grid->dv,0,y-1,x))   / grid->dy
                  + (DARRAY(grid->pnuv,0,y,x) - DARRAY(grid->pnuv,0,y,x-1)) / grid->dx
                );

        // Time corrector
        # pragma omp for
        for ( int_t y=0; y<grid->length.y; y++ )
            for ( int_t x=0; x<grid->length.x; x++ )
                DARRAY(grid->pn,2,y,x) =
                    (DARRAY(grid->pn,0,y,x) + DARRAY(grid->pn,1,y,x)) / 2
                  - (grid->dt / 2) * (
                    (DARRAY(grid->pnu,1,y,x) - DARRAY(grid->pnu,1,y,x-1)) / grid->dx
                  + (DARRAY(grid->pnv,1,y,x) - DARRAY(grid->pnv,1,y-1,x)) / grid->dy
                );
    }

    rank_time[integrate_time_step_index] += omp_get_wtime() - start;
}


void update_time_step(grid_t *grid) {
    real_t start = omp_get_wtime();
    
    # pragma omp parallel
    {
        # pragma omp for
        // Including exterior (and halo, although it will be overwritten)
        for (int_t y = -grid->border; y < grid->length.y + grid->border; y++)
            for (int_t x = -grid->border; x < grid->length.x + grid->border; x++)
                DARRAY(grid->n, 1, y, x) = DARRAY(grid->pn, 2, y, x) / P;

        # pragma omp for
        for (int_t y = -grid->border; y < grid->length.y + grid->border; y++)
            for (int_t x = -grid->border; x < grid->length.x + grid->border; x++)
                DARRAY(grid->u, 1, y, x) =
                    DARRAY(grid->pnu, 2, y, x) / DARRAY(grid->pn, 2, y, x);

        # pragma omp for
        for (int_t y = -grid->border; y < grid->length.y + grid->border; y++)
            for (int_t x = -grid->border; x < grid->length.x + grid->border; x++)
                DARRAY(grid->v, 1, y, x) =
                    DARRAY(grid->pnv, 2, y, x) / DARRAY(grid->pn, 2, y, x);
    }

    rank_time[update_time_step_index] += omp_get_wtime() - start;
}


// Find and print output
void print_output(real_t total_rank_time) {
    // Find output

    // Runtime profiling
    real_t
        // rank_reciprocals[n_profile_functions],
        reciprocals[n_profile_functions],
        harmonic_means[n_profile_functions],
        time[n_profile_functions],
        arithmetic_mean_time[n_profile_functions],
        arithmetic_share[n_profile_functions];

    // for (int_t i = 0; i < n_profile_functions; i++) {
    //     rank_reciprocals[i] = total_rank_time / rank_time[i];
    // }

    // MPI_Reduce(rank_reciprocals, reciprocals, n_profile_functions, MPI_DOUBLE,
    //            MPI_SUM, 0, MPI_COMM_WORLD);

    MPI_Reduce(rank_time, time, n_profile_functions, MPI_DOUBLE, MPI_SUM, 0,
               MPI_COMM_WORLD);

    real_t total_time;

    MPI_Reduce(&total_rank_time, &total_time, 1, MPI_DOUBLE, MPI_MAX, 0,
               MPI_COMM_WORLD);

    real_t total_rank_time_without_wait = total_rank_time - rank_time[wait_index];

    for (int_t i = 0; i < n_profile_functions; i++) {
        // harmonic_means[i] = 100 * size / reciprocals[i];
        arithmetic_mean_time[i] = time[i] / size;
        arithmetic_share[i] = 100 * arithmetic_mean_time[i] / total_time;
    }

    MPI_Reduce(&n_rank_points, &n_total_points, 1, MPI_INT64_T, MPI_SUM, 0,
               MPI_COMM_WORLD);

    MPI_Reduce(&n_rank_child_points, &n_total_child_points, 1, MPI_INT64_T,
               MPI_SUM, 0, MPI_COMM_WORLD);

    // Print output
    if (rank == 0) {
        printf("Y: %" PRId64 "\n", Y);
        printf("X: %" PRId64 "\n", X);
        printf("Max refinement: %" PRId64 "\n", max_refinement);
        printf("MPI_size: %d\n", size);
        printf("OpenMP_max_threads: %d\n", omp_get_max_threads());

        for (int_t i = 0; i < n_profile_functions; i++) {
            printf("%s mean time:\t%.2f\n",
                   profile_function_names[i], arithmetic_mean_time[i]);

            printf("%s share:\t%.2f%%\n", profile_function_names[i],
                   arithmetic_share[i]);

            // printf("%s harmonic share:\t%.2f%%\n", profile_function_names[i],
            //        harmonic_means[i]);
        }

        printf("Total points: %" PRId64 "\n", n_total_points);
        printf("Total child points: %" PRId64 "\n", n_total_child_points);

        if (measure_load_balance) {
            printf("Total load points: %" PRId64 "\n", n_total_load_points);

            // We define load balance as a number between 0 and 1.
            // A value of 1 indicates a perfect load balance.
            real_t load_balance = (real_t)n_total_points / n_total_load_points;
            printf("Load balance: %f\n", load_balance);
        }

        printf("Total time:\t%.2f\n", total_time);
    }

    MPI_Barrier(cart);

    printf("Rank %d: %" PRId64 " points, dims_y: %d, dims_x: %d\n", rank,
           n_rank_points, dims[0], dims[1]);

    MPI_Barrier(cart);

    printf("Rank %d: rank border time: %.2f\n", rank,
           rank_time[rank_border_index]);

    MPI_Barrier(cart);

    printf("Rank %d: rank wait time: %.2f\n", rank, rank_time[wait_index]);

    MPI_Barrier(cart);

    printf("Rank %d: Total rank time without wait: %.2f\n", rank, total_rank_time_without_wait);
}


void update_load_balance_measure() {
    // The number of points for a given rank in a given iteration for
    // root
    int_t n_rank_root_iter_points =
        n_rank_points - n_rank_points_last_root_iter;

    // Update for the next iteration
    n_rank_points_last_root_iter = n_rank_points;

    // The maximum number of points for a single rank in a given
    // snapshot
    int_t n_max_root_iter_points;

    // Find the largest number of points for a single rank in this
    // snapshot.
    MPI_Reduce(&n_rank_root_iter_points, &n_max_root_iter_points, 1,
                MPI_INT64_T, MPI_MAX, 0, MPI_COMM_WORLD);

    if (rank == 0) {
        // The cost of an imbalanced load is equal to the cost of a
        // perfectly balanced load where all ranks do the same amount
        // of work as the hardest working rank in an imbalanced load.
        n_total_load_points += n_max_root_iter_points * size;
    }
}
